#!/usr/bin/env python
'''Config file for galaxy diver.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import os

########################################################################

DIR = os.path.dirname( __file__ )
'''Where most of the source code is.'''

DATA_DIR = os.path.join( DIR, 'data' )
'''Where the data is store.'''
